"""
This module contains classes that describe different optical sensors in the system.

This includes scoring cameras and wavefront sensors, and their respective detectors.
"""

from dataclasses import dataclass, field

from .aberration import Aberration
from .base import Referenceable, Coordinates
from .image import Image
from .source import Source

__all__ = ['Detector', 'ScoringCamera', 'WavefrontSensor', 'ShackHartmann', 'Pyramid']


@dataclass(kw_only=True)
class Detector(Referenceable):
    """Contains data regarding a detector in the system.
    Detectors are a part of wavefront sensors and scoring cameras."""

    type: str = None
    "Identifies the type of detector, such as ``'CMOS'`` or ``'CCD'``. (dimensionless quantity)"

    sampling_technique: str = None
    """Identifies the sampling technique, for example ``'Single Reset Read'``, ``'Fowler'``, ``'Double Correlated'``,
     ``'Uncorrelated'``. (dimensionless quantity)"""

    shutter_type: str = None
    "Identifies the shutter type, typically ``'Rolling'`` or ``'Global'``. (dimensionless quantity)"

    flat_field: Image = None
    """Inverse of the detector pixel-to-pixel sensitivity
    (Dimensions :math:`h \\times w`, dimensionless quantity, using data type flt)"""

    readout_noise: float = None
    r'Readout noise. (in e\ :math:`^-` s\ :math:`^{-1}` pix\ :math:`^{-1}` units)'

    pixel_intensities: Image = None
    """Intensity detected in each pixel, for each data frame. This is a sequence of :math:`t` images, each spanning 
    :math:`x` pixels horizontally and :math:`y` pixels vertically.
    (Dimensions :math:`t \\times h \\times w`, in ADU units, using data type flt)"""

    integration_time: float = None
    'TODO: Integration time. (in s units)'

    coadds: int = None
    'Number of frame co-additions.  (in count units)'

    dark: Image = None
    """Intensity detected in each pixel when there is no light being observed. This is an image spanning 
    :math:`x` pixels horizontally and :math:`y` pixels vertically.
    (Dimensions :math:`h \\times w`, in ADU units, using data type flt)"""

    weight_map: Image = None
    """Pixel weight map, where each detector pixel is associated with a number that represents its relative value.
    Summing up to 1. (Dimensions :math:`h \\times w`, dimensionless quantity, using data type flt)"""

    quantum_efficiency: float = None
    'TODO: Quantum efficiency. (dimensionless quantity)'

    pixel_scale: float = None
    r'TODO: Pixel scale. (in rad pix\ :math:`^{-1}` units)'

    binning: int = None
    'TODO: Binning. (in count units)'

    bandwidth: float = None
    'TODO: Spectral bandwidth. (in m units)'

    transmission_wavelength: list[float] = field(default_factory=list)
    'List of wavelengths that describe a transmission profile. (in m units)'

    transmission: list[float] = field(default_factory=list)
    'List of transmission percentages that describe a transmission profile. (dimensionless quantity)'

    sky_background: Image = None
    """detector pixel intensities from a source-less direction in the sky
    (Dimensions :math:`h \\times w`, in ADU units, using data type flt)"""

    gain: float = None
    r'TODO: Gain. (in e\ :math:`^-` units)'

    excess_noise: float = None
    r'TODO: Excess noise. (in e\ :math:`^-` units)'

    filter: str = None
    'Name of filter in use. (dimensionless quantity)'

    bad_pixel_map: Image = None
    """Binary image which identifies the bad pixels. Pixels identified with 1 are considered bad, while 0 is considered
    normal.  (Dimensions :math:`h \\times w`, dimensionless quantity, using data type int)"""

    dynamic_range: float = None
    """Ratio of the maximum signal that can be integrated to the r.m.s. noise floor.
    If this ratio is R, then dynamic range in decibels is 20 log R. (in dB units)"""

    readout_rate: float = None
    r'Inverse of the time required to digitize a single pixel (in px s\ :math:`^{-1}` units)'

    frame_rate: float = None
    r"""Inverse of the time needed for the detector to acquire an image and then completely read it out.
    (in frame s\ :math:`^{-1}` units)"""

    transformation_matrix: Image = None
    r"""Matrix that defines 2-dimensional affine transformations over time (:math:`t`) using homogeneous coordinates.
    Any combination of translation, reflection, scale, rotation and shearing can be described via a single
    :math:`3 \times 3` matrix :math:`M` such that :math:`P' = MP`, where :math:`P` is a
    :math:`\begin{bmatrix}x & y & 1 \end{bmatrix}` vector (with :math:`x` and :math:`y` being the original horizontal 
    and vertical coordinates, respectively) and :math:`P'` is a :math:`\begin{bmatrix}x' & y' & 1 \end{bmatrix}`, where
    :math:`x'` and :math:`y'` are the transformed coordinates. All geometry information must be described relative to
    the same reference origin point, from which transformations may occur.
    (Dimensions :math:`3 \times 3 \times t`, dimensionless quantity, using data type flt)"""


@dataclass(kw_only=True)
class ScoringCamera(Referenceable):
    """Contains data regarding a scoring camera in the system."""

    pupil_mask: Image = None
    """Binary image that describes the shape of the pupil. A 1 indicates the presence of the pupil, while a 0 indicates
    the opposite. (Dimensions :math:`h \\times w`, dimensionless quantity, using data type int)"""

    wavelength: float = None
    """'Observation wavelength  (in m units)'"""

    transformation_matrix: Image = None
    r"""Matrix that defines 2-dimensional affine transformations over time (:math:`t`) using homogeneous coordinates.
    Any combination of translation, reflection, scale, rotation and shearing can be described via a single
    :math:`3 \times 3` matrix :math:`M` such that :math:`P' = MP`, where :math:`P` is a
    :math:`\begin{bmatrix}x & y & 1 \end{bmatrix}` vector (with :math:`x` and :math:`y` being the original horizontal 
    and vertical coordinates, respectively) and :math:`P'` is a :math:`\begin{bmatrix}x' & y' & 1 \end{bmatrix}`, where
    :math:`x'` and :math:`y'` are the transformed coordinates. All geometry information must be described relative to
    the same reference origin point, from which transformations may occur.
    (Dimensions :math:`3 \times 3 \times t`, dimensionless quantity, using data type flt)"""

    detector: Detector = None
    aberration: Aberration = None


@dataclass(kw_only=True)
class WavefrontSensor(Referenceable):
    """Abstract class that contains data related to one wavefront sensor in the system."""

    source: Source

    dimensions: int
    """Number of dimensions being measured by each subaperture. (in count units)"""

    n_valid_subapertures: int
    'Number of valid subapertures (must coincide with `subaperture_mask` data). (in count units)'

    measurements: Image = None
    """Measurements from the sensor over time. Each of its :math:`s_v` subapertures is able to measure in :math:`d`
    dimensions.  (Dimensions :math:`t \\times d \\times s_v`, in user defined units, using data type flt)"""

    ref_measurements: Image = None
    'Reference measurements. (Dimensions :math:`d \\times s_v`, in user defined units, using data type flt)'

    subaperture_mask: Image = None
    r"""Representation of the subaperture grid, where the cells corresponding to invalid subapertures are marked as
    :math:`-1` and the cells corresponding to valid subapertures contain their respective index in the sequence of
    valid subapertures (using zero-based numbering, that is, the initial element is assigned the index 0).
    (Dimensions :math:`s \times s`, dimensionless quantity, using data type int)"""

    mask_offsets: list[Coordinates] = field(default_factory=list)
    """List of horizontal/vertical offsets in detector pixels, represented as `Coordinates`.
    Each offset defines the lowest horizontal/vertical position occupied by the respective mask. (in pix units)"""

    subaperture_size: float = None
    'Size of each subaperture in detector pixels. (in pix units)'

    subaperture_intensities: Image = None
    """Detected average intensity (flux) of each of the :math:`s_v` valid subapertures, over :math:`t` time.
    (Dimensions :math:`t \\times s_v`, in ADU units, using data type flt)"""

    wavelength: float = None
    'Wavelength being sensed. (in m units)'

    optical_gain: Image = None
    'WFS optical gain over time. (Dimensions :math:`t`, dimensionless quantity, using data type flt)'

    transformation_matrix: Image = None
    r"""Matrix that defines 2-dimensional affine transformations over time (:math:`t`) using homogeneous coordinates.
    Any combination of translation, reflection, scale, rotation and shearing can be described via a single
    :math:`3 \times 3` matrix :math:`M` such that :math:`P' = MP`, where :math:`P` is a
    :math:`\begin{bmatrix}x & y & 1 \end{bmatrix}` vector (with :math:`x` and :math:`y` being the original horizontal 
    and vertical coordinates, respectively) and :math:`P'` is a :math:`\begin{bmatrix}x' & y' & 1 \end{bmatrix}`, where
    :math:`x'` and :math:`y'` are the transformed coordinates. All geometry information must be described relative to
    the same reference origin point, from which transformations may occur.
    (Dimensions :math:`3 \times 3 \times t`, dimensionless quantity, using data type flt)"""

    detector: Detector = None
    aberration: Aberration = None
    non_common_path_aberration: Aberration = None

    def __post_init__(self):
        if self.__class__ == WavefrontSensor:
            raise TypeError("Cannot instantiate abstract class.")


@dataclass(kw_only=True)
class ShackHartmann(WavefrontSensor):
    """Contains data related to one Shack-Hartmann wavefront sensor used by the system."""
    dimensions: int = field(init=False, default=2)
    """Number of dimensions being measured by each subaperture. For `ShackHartmann` this must be equal to 2
    (horizontal and vertical offset). (in count units)"""

    centroiding_algorithm: str = None
    'Name of the centroiding algorithm used. (dimensionless quantity)'

    centroid_gains: Image = None
    """Centroid gain factors for each of :math:`s_v` subapertures and :math:`d` dimensions.
    (Dimensions :math:`d \\times s_v`, dimensionless quantity, using data type flt)"""

    spot_fwhm: Image = None
    """Spot full width half maximum for each of :math:`s_v` subapertures and  :math:`d` dimensions.
    (Dimensions :math:`d \\times s_v`, in arcsec units, using data type flt)"""


@dataclass(kw_only=True)
class Pyramid(WavefrontSensor):
    """Contains data related to one Pyramid wavefront sensor used by the system."""
    dimensions: int = 4
    """Number of dimensions being measured by each subaperture. For `Pyramid` this may be equal to 2 (if the signals are
    interpreted as horizontal and vertical offsets), 1 (if the subapertures overlap and are interpreted as a single
    signal) or as the number of sides of the pyramid (that is, `n_sides` signals). (in count units)"""

    n_sides: int
    'Number of pyramid sides (typically 4).  (in count units)'

    modulation: float = None
    'Modulation amplitude.  (in m units)'



"""
This module contains classes and functions that enable reading AOT FITS files.
"""

import re
import warnings
from datetime import datetime

import numpy as np
from astropy.io import fits

import aotpy
from . import _keywords as kw
from .utils import FITSURLImage, FITSFileImage, image_from_hdu, keyword_is_relevant, metadatum_from_card
from ..base import SystemReader

_reference_pattern = re.compile(r'([^<]+)<(.+)>(\d+)?')


def _type_matches(aot_format: str, fits_type: str) -> bool:
    try:
        return aot_format == kw.fits_type_to_aot(fits_type)
    except ValueError:
        return False


def _convert_row(columns: fits.ColDefs, row: fits.FITS_record, fields: kw.AOTFieldDict) -> dict:
    res = {}
    for name, field in fields.items():
        try:
            value = row[name]
        except KeyError:
            value = None

        match field.format:
            case kw.STRING_FORMAT:
                if value == '':
                    value = None
            case kw.LIST_FORMAT:
                if value is None:
                    value = []
                value = [None if np.isnan(v) else v for v in value]
            case kw.FLOAT_FORMAT:
                if np.isnan(value):
                    value = None
            case kw.INTEGER_FORMAT:
                if value is not None:
                    col = columns[name]
                    if col.null and value == col.null:
                        value = None
            case _:
                raise RuntimeError  # This should never happen
        if value is None and field.mandatory:
            # This will never trigger in list fields, but it's ok because we don't have mandatory list fields
            raise ValueError(f"Found null value in mandatory field '{name}'")
        res[name] = value
    return res


def read_system_from_fits(filename: str, extra_data: bool = False, **kwargs) -> aotpy.AOSystem:
    """
    Get `AOSystem` from FITS file specified by `filename`.

    Parameters
    ----------
    filename
        Path to file to be read into an `AOSystem`.
    extra_data : default = False
        Whether it is expected that the file contains some data that does not fit the AOT standard. If `extra_data` is
        not `True`, user will be warned if extra data is detected.
    **kwargs
        Keyword arguments passed on as options to the file handling function.
    """
    r = FITSReader(filename, extra_data=extra_data, **kwargs)
    return r.get_system()


class FITSReader(SystemReader):
    def _initialize_data(self) -> None:
        """
        Initialize data structures necessary for reading the file.
        """
        self._images: dict[str, list] = {}
        self._time: dict[str, list] = {}
        self._aberrations: dict[str, list] = {}
        self._telescopes: dict[str, list] = {}
        self._sources: dict[str, list] = {}
        self._detectors: dict[str, list] = {}
        self._wfss: dict[str, list] = {}
        self._wfcs: dict[str, list] = {}

        self._table_to_dict = {
            kw.TIME_TABLE: self._time,
            kw.ABERRATIONS_TABLE: self._aberrations,
            kw.TELESCOPES_TABLE: self._telescopes,
            kw.SOURCES_TABLE: self._sources,
            kw.DETECTORS_TABLE: self._detectors,
            kw.WAVEFRONT_SENSORS_TABLE: self._wfss,
            kw.WAVEFRONT_CORRECTORS_TABLE: self._wfcs
        }

        self._extra_hdus: fits.HDUList = fits.HDUList()
        self._extra_columns: dict[str, list[fits.Column]] = {}
        self._extra_objects: list[aotpy.Referenceable] = []

    def _read(self, **kwargs) -> tuple[aotpy.AOSystem, list]:
        with fits.open(self._filename, **kwargs) as hdus:
            self._primary_header: fits.Header = hdus[0].header
            self._system: aotpy.AOSystem = self._check_header()

            if hdus[0].data is not None:
                raise ValueError('Primary HDU must have no data.')
            for table in kw.MANDATORY_TABLE_SET:
                if table not in hdus:
                    raise ValueError(f"Missing mandatory binary table '{table}'.")

            table_count = {table: 0 for table in kw.TABLE_SET}
            # Skip PrimaryHDU
            for hdu in hdus[1:]:
                if hdu.name in table_count:
                    table_count[hdu.name] += 1
                else:
                    if hdu.is_image:
                        if not hdu.name:
                            raise ValueError('All image extensions in file must have a name.')
                        if hdu.name in self._images:
                            raise ValueError(f"Image name '{hdu.name}' appears repeated in file.")
                        if hdu.data is None:
                            warnings.warn(f"Image HDU '{hdu.name}' was ignored for having no data.")
                        self._images[hdu.name] = [image_from_hdu(hdu), False]
                    else:
                        self._extra_hdus.append(hdu)
            if self._extra_hdus and not self._extra_data_flag:
                warnings.warn(f"""File contains non-AOT HDUs that were ignored: """
                              f"""{', '.join([f"'{x.name}'" for x in self._extra_hdus])}""")

            for name, count in table_count.items():
                if name in kw.MANDATORY_TABLE_SET and count != 1:
                    raise ValueError(f"Mandatory table '{name}' must appear exactly once in file.")
                if count > 1:
                    raise ValueError(f"Secondary table '{name}' cannot appear repeated in file.")
            seq = [table for table in kw.TABLE_SEQUENCE if table_count[table] == 1]
            if seq != [hdu.name for hdu, _ in zip(hdus[1:], seq)]:
                warnings.warn('File does not follow the standard AOT table sequence.')

            self._handle_time(hdus)

            for tup in self._images.values():
                image = tup[0]
                image.time = self._handle_reference(image._time, kw.TIME_TABLE)

            self._handle_atmosphere(hdus)
            self._handle_aberrations(hdus)
            self._handle_telescopes(hdus)
            self._handle_sources(hdus)
            self._handle_detectors(hdus)
            self._handle_scoring_cameras(hdus)
            self._handle_wavefront_sensors(hdus)
            self._handle_wavefront_correctors(hdus)
            self._handle_loops(hdus)

            self._check_usage()
            aux = [self._extra_hdus, self._extra_columns, self._extra_objects]
            extra = aux if [x for x in aux if x] else None  # extra is None if everything is empty
            return self._system, extra

    def _check_header(self) -> aotpy.AOSystem:
        if kw.AOT_VERSION not in self._primary_header:
            raise ValueError(f"File is not in the AOT format or the mandatory '{kw.AOT_VERSION}' keyword is missing")
            # TODO do something with the version
        try:
            if self._primary_header[kw.AOT_TIMESYS] != kw.AOT_TIMESYS_UTC:
                raise ValueError(f"Keyword '{kw.AOT_TIMESYS}' must have the value '{kw.AOT_TIMESYS_UTC}'")
        except KeyError:
            raise ValueError(f"Mandatory keyword '{kw.AOT_TIMESYS}' is missing") from None

        ao_mode = None
        try:
            aux = self._primary_header[kw.AOT_AO_MODE]
            if aux in kw.AOT_AO_MODE_SET:
                ao_mode = aux
            else:
                warnings.warn(f"Ignored unrecognized value '{aux}' for keyword '{kw.AOT_AO_MODE}'. "
                              f"Value should be one of: {str(kw.AOT_AO_MODE_SET)[1:-1]}")
        except KeyError:
            warnings.warn(f"Mandatory keyword '{kw.AOT_AO_MODE}' is missing")

        beg = None
        try:
            aux = self._primary_header[kw.AOT_DATE_BEG]
            if aux:
                try:
                    beg = datetime.fromisoformat(aux)
                except (ValueError, TypeError):
                    warnings.warn(f"'{kw.AOT_DATE_BEG}' keyword not properly formatted.")
        except KeyError:
            pass

        end = None
        try:
            aux = self._primary_header[kw.AOT_DATE_END]
            if aux:
                try:
                    end = datetime.fromisoformat(aux)
                except (ValueError, TypeError):
                    warnings.warn(f"'{kw.AOT_DATE_END}' keyword not properly formatted.")
        except KeyError:
            pass

        try:
            name = self._primary_header[kw.AOT_SYSTEM_NAME]
        except KeyError:
            name = None

        try:
            strehl_ratio = self._primary_header[kw.AOT_STREHL_RATIO]
        except KeyError:
            strehl_ratio = None

        try:
            temporal_error = self._primary_header[kw.AOT_TEMPORAL_ERROR]
        except KeyError:
            temporal_error = None

        try:
            config = self._primary_header[kw.AOT_CONFIG]
        except KeyError:
            config = None

        metadata = []
        for card in self._primary_header.cards:
            if card.keyword not in kw.AOT_HEADER_SET and keyword_is_relevant(card.keyword):
                metadata.append(metadatum_from_card(card))

        return aotpy.AOSystem(ao_mode=ao_mode, date_beginning=beg, date_end=end, name=name, strehl_ratio=strehl_ratio,
                              temporal_error=temporal_error, config=config, metadata=metadata)

    def _check_bintable(self, hdus: fits.HDUList, table_name: str):
        fields = kw.TABLE_FIELDS[table_name]
        table = hdus[table_name]
        if not isinstance(table, fits.BinTableHDU):
            raise ValueError(f"HDU '{table.name}' must be a binary table")

        field_count = {field: 0 for field in fields}
        for col in table.columns.columns:
            if col.name in field_count:
                field_count[col.name] += 1
                field = fields[col.name]
                if not _type_matches(field.format, col.format):
                    raise ValueError(f"Column '{col.name}' in table '{table_name}' must be a '{field.format}'"
                                     f" field (found '{col.format}')")
                if col.unit != field.unit:
                    warnings.warn(f"Column '{col.name}' in table '{table_name}' does not have a standard unit. "
                                  f"Found '{col.unit}', expected '{field.unit}'")
                if field.unique:
                    u = table.data[col.name]
                    if len(np.unique(u)) != len(u):
                        raise ValueError(
                            f"Column '{col.name}' in table '{table_name}' must have unique values.")
            else:
                self._extra_columns.setdefault(table_name, []).append(col)
        if table_name in self._extra_columns and not self._extra_data_flag:
            warnings.warn(f"""Table '{table_name}' contains non-AOT columns that were ignored: """
                          f"""{', '.join([f"'{x.name}'" for x in self._extra_columns[table_name]])}""")

        for name, count in field_count.items():
            if count < 1:
                warnings.warn(f"Column '{name}' missing in table '{table_name}'.")
            if count > 1:
                raise ValueError(f"Column '{name}' cannot appear repeated in table '{table_name}'.")

        seq = [field_name for field_name, count in field_count.items() if count > 0]
        if seq != [col.name for col, _ in zip(table.columns.columns, seq)]:
            warnings.warn(f"Non-AOT column sequence in table '{table_name}'.")

    def _handle_image(self, ref: str):
        if ref is None:
            return None
        fullmatch = _reference_pattern.fullmatch(ref)
        if fullmatch is None:
            warnings.warn(f"Image reference '{ref}' was ignored: not properly formatted.")
            return None
        prefix, name, index = fullmatch.groups()

        match prefix:
            case kw.INTERNAL_REFERENCE:
                try:
                    aux = self._images[name]
                    aux[1] = True
                    return aux[0]
                except KeyError:
                    warnings.warn(f"Could not find internal image referenced by '{ref}'.")
                    return None
            case kw.FILE_REFERENCE | kw.URL_REFERENCE:
                if index is not None:
                    try:
                        index = int(index)
                    except ValueError:
                        warnings.warn(f"Index in file reference '{ref}' was ignored: not properly formatted.")
                        index = None
                if prefix == kw.FILE_REFERENCE:
                    image = FITSFileImage(name, index)
                else:
                    image = FITSURLImage(name, index)

                image.time = self._handle_reference(image._time, kw.TIME_TABLE)
                return image
            case _:
                warnings.warn(f"Reference '{ref}' was ignored: expected an image reference.")
                return None

    def _handle_reference(self, ref: str, table: str):
        if ref is None:
            return None
        fullmatch = _reference_pattern.fullmatch(ref)
        if fullmatch is None:
            warnings.warn(f"Row reference '{ref}' was ignored: not properly formatted.")
            return None
        prefix, name, index = fullmatch.groups()

        if prefix != kw.ROW_REFERENCE:
            warnings.warn(f"Reference '{ref}' was ignored: expected a row reference.")
            return None

        d = self._table_to_dict[table]
        try:
            aux = d[name]
            aux[1] = True
            return aux[0]
        except KeyError:
            warnings.warn(f"Could not find row referenced by '{ref}'. Ignoring reference.")
            return None

    def _check_usage(self):
        # We don't need to check the usage of atmosphere parameters, sources, scoring cameras, wavefront sensors,
        # wavefront correctors or loops, since they are always referenced by the AOSystem class
        a = [(self._images, 'image extensions'),
             (self._time, 'time rows'),
             (self._aberrations, 'aberration rows'),
             (self._telescopes, 'telescope rows'),
             (self._detectors, 'detector rows')]

        for d, name in a:
            unused = {k: v[0] for k, v in d.items() if not v[1]}
            if unused:
                self._extra_objects.extend(unused.values())
                if not self._extra_data_flag:
                    warnings.warn(f"""File contains some {name} were ignored for never being referenced: """
                                  f"""{', '.join([f"'{x}'" for x in unused])}""")

    def _handle_time(self, hdus: fits.HDUList):
        self._check_bintable(hdus, kw.TIME_TABLE)

        table = hdus[kw.TIME_TABLE]
        columns = table.columns
        for row in table.data:
            data = _convert_row(columns, row, kw.TIME_FIELDS)

            self._time[data[kw.REFERENCE_UID]] = [aotpy.Time(
                uid=data[kw.REFERENCE_UID],
                timestamps=data[kw.TIME_TIMESTAMPS],
                frame_numbers=data[kw.TIME_FRAME_NUMBERS]
            ), False]

    def _handle_atmosphere(self, hdus: fits.HDUList):
        self._check_bintable(hdus, kw.ATMOSPHERIC_PARAMETERS_TABLE)

        table = hdus[kw.ATMOSPHERIC_PARAMETERS_TABLE]
        columns = table.columns
        for row in table.data:
            data = _convert_row(columns, row, kw.ATMOSPHERIC_PARAMETERS_FIELDS)

            self._system.atmosphere_params.append(
                aotpy.AtmosphericParameters(
                    uid=data[kw.REFERENCE_UID],
                    wavelength=data[kw.ATMOSPHERIC_PARAMETERS_WAVELENGTH],
                    time=self._handle_reference(data[kw.TIME_REFERENCE], kw.TIME_TABLE),
                    r0=data[kw.ATMOSPHERIC_PARAMETERS_R0],
                    fwhm=data[kw.ATMOSPHERIC_PARAMETERS_FWHM],
                    tau0=data[kw.ATMOSPHERIC_PARAMETERS_TAU0],
                    theta0=data[kw.ATMOSPHERIC_PARAMETERS_THETA0],
                    layers_weight=self._handle_image(data[kw.ATMOSPHERIC_PARAMETERS_LAYERS_WEIGHT]),
                    layers_height=self._handle_image(data[kw.ATMOSPHERIC_PARAMETERS_LAYERS_HEIGHT]),
                    layers_l0=self._handle_image(data[kw.ATMOSPHERIC_PARAMETERS_LAYERS_L0]),
                    layers_wind_speed=self._handle_image(data[kw.ATMOSPHERIC_PARAMETERS_LAYERS_WIND_SPEED]),
                    layers_wind_direction=self._handle_image(data[kw.ATMOSPHERIC_PARAMETERS_LAYERS_WIND_DIRECTION]),
                    transformation_matrix=self._handle_image(data[kw.TRANSFORMATION_MATRIX])
                )
            )

    def _handle_aberrations(self, hdus: fits.HDUList):
        self._check_bintable(hdus, kw.ABERRATIONS_TABLE)

        table = hdus[kw.ABERRATIONS_TABLE]
        columns = table.columns
        for row in table.data:
            data = _convert_row(columns, row, kw.ABERRATION_FIELDS)

            self._aberrations[data[kw.REFERENCE_UID]] = [aotpy.Aberration(
                uid=data[kw.REFERENCE_UID],
                modes=self._handle_image(data[kw.ABERRATION_MODES]),
                coefficients=self._handle_image(data[kw.ABERRATION_COEFFICIENTS]),
                offsets=[aotpy.Coordinates(x, y)
                         for x, y in zip(data[kw.ABERRATION_X_OFFSETS], data[kw.ABERRATION_Y_OFFSETS])],
            ), False]

    def _handle_telescopes(self, hdus: fits.HDUList):
        self._check_bintable(hdus, kw.TELESCOPES_TABLE)

        table = hdus[kw.TELESCOPES_TABLE]
        columns = table.columns
        for row in table.data:
            data = _convert_row(columns, row, kw.TELESCOPE_FIELDS)

            uid = data[kw.REFERENCE_UID]
            t = data[kw.TELESCOPE_TYPE]
            if t == kw.TELESCOPE_TYPE_MAIN:
                if self._system.main_telescope is not None:
                    raise ValueError(f"There must only be one telescope of type '{kw.TELESCOPE_TYPE_MAIN}'.")
                tel = aotpy.MainTelescope(uid)
                self._system.main_telescope = tel
                self._telescopes[uid] = [tel, True]  # The main telescope is always referenced by the AOSystem
            elif t == kw.TELESCOPE_TYPE_LLT:
                tel = aotpy.LaserLaunchTelescope(uid)
                self._telescopes[uid] = [tel, False]
            else:
                warnings.warn(f"Skipped telescope {uid}: unknown type '{t}'.")
                continue

            tel.latitude = data[kw.TELESCOPE_LATITUDE]
            tel.longitude = data[kw.TELESCOPE_LONGITUDE]
            tel.elevation = data[kw.TELESCOPE_ELEVATION]
            tel.azimuth = data[kw.TELESCOPE_AZIMUTH]
            tel.parallactic = data[kw.TELESCOPE_PARALLACTIC]
            tel.pupil_mask = self._handle_image(data[kw.TELESCOPE_PUPIL_MASK])
            tel.pupil_angle = data[kw.TELESCOPE_PUPIL_ANGLE]
            tel.enclosing_diameter = data[kw.TELESCOPE_ENCLOSING_D]
            tel.inscribed_diameter = data[kw.TELESCOPE_INSCRIBED_D]
            tel.obstruction_diameter = data[kw.TELESCOPE_OBSTRUCTION_D]

            t = data[kw.TELESCOPE_SEGMENTS_TYPE]
            if t == kw.TELESCOPE_SEGMENT_TYPE_MONOLITHIC:
                seg = aotpy.Monolithic()
            else:
                if t == kw.TELESCOPE_SEGMENT_TYPE_CIRCLE:
                    seg = aotpy.CircularSegments()
                elif t == kw.TELESCOPE_SEGMENT_TYPE_HEXAGON:
                    seg = aotpy.HexagonalSegments()
                else:
                    warnings.warn(f"Ignored unknown segment type '{t}'.")
                    seg = aotpy.Monolithic()
                seg.size = data[kw.TELESCOPE_SEGMENTS_SIZE]
                seg.coordinates = [aotpy.Coordinates(x, y)
                                   for x, y in zip(data[kw.TELESCOPE_SEGMENTS_X], data[kw.TELESCOPE_SEGMENTS_Y])]

            tel.segments = seg
            tel.transformation_matrix = self._handle_image(data[kw.TRANSFORMATION_MATRIX])
            tel.aberration = self._handle_reference(data[kw.ABERRATION_REFERENCE], kw.ABERRATIONS_TABLE)

    def _handle_sources(self, hdus: fits.HDUList):
        self._check_bintable(hdus, kw.SOURCES_TABLE)

        table = hdus[kw.SOURCES_TABLE]
        existing_types = table.data['TYPE']
        if kw.SOURCE_TYPE_SODIUM_LASER_GUIDE_STAR in existing_types:
            if kw.SOURCES_SODIUM_LGS_TABLE in hdus:
                self._check_bintable(hdus, kw.SOURCES_SODIUM_LGS_TABLE)
            else:
                raise ValueError(f"Missing table '{kw.SOURCES_SODIUM_LGS_TABLE}' must exist when "
                                 f"'{kw.SOURCE_TYPE_SODIUM_LASER_GUIDE_STAR}' type sources exist")
        if kw.SOURCE_TYPE_RAYLEIGH_LASER_GUIDE_STAR in existing_types:
            if kw.SOURCES_RAYLEIGH_LGS_TABLE in hdus:
                self._check_bintable(hdus, kw.SOURCES_RAYLEIGH_LGS_TABLE)
            else:
                raise ValueError(f"Missing table '{kw.SOURCES_SODIUM_LGS_TABLE}' must exist when "
                                 f"'{kw.SOURCE_TYPE_RAYLEIGH_LASER_GUIDE_STAR}' type sources exist")

        columns = table.columns
        for row in table.data:
            data = _convert_row(columns, row, kw.SOURCE_FIELDS)

            uid = data[kw.REFERENCE_UID]
            t = data[kw.SOURCE_TYPE]
            if t == kw.SOURCE_TYPE_SCIENCE_STAR:
                src = aotpy.ScienceStar(uid)
            elif t == kw.SOURCE_TYPE_NATURAL_GUIDE_STAR:
                src = aotpy.NaturalGuideStar(uid)
            elif t == kw.SOURCE_TYPE_SODIUM_LASER_GUIDE_STAR:
                # Try to find uid in secondary table
                f = next((x for x in hdus[kw.SOURCES_SODIUM_LGS_TABLE].data if x[kw.REFERENCE_UID] == uid), None)
                if f is None:
                    raise ValueError(f"Source '{uid}' not found in table '{kw.SOURCES_SODIUM_LGS_TABLE}' even"
                                     f" though it is of type '{t}'")

                other_data = _convert_row(hdus[kw.SOURCES_SODIUM_LGS_TABLE].columns, f, kw.SOURCE_SODIUM_LGS_FIELDS)
                src = aotpy.SodiumLaserGuideStar(
                    uid=uid,
                    height=other_data[kw.SOURCE_SODIUM_LGS_HEIGHT],
                    profile=self._handle_image(other_data[kw.SOURCE_SODIUM_LGS_PROFILE]),
                    altitudes=other_data[kw.SOURCE_SODIUM_LGS_ALTITUDES],
                    laser_launch_telescope=self._handle_reference(other_data[kw.LASER_LAUNCH_TELESCOPE_REFERENCE],
                                                                  kw.TELESCOPES_TABLE)
                )
            elif t == kw.SOURCE_TYPE_RAYLEIGH_LASER_GUIDE_STAR:
                # Try to find uid in secondary table
                f = next((x for x in hdus[kw.SOURCES_RAYLEIGH_LGS_TABLE].data if x[kw.REFERENCE_UID] == uid), None)
                if f is None:
                    raise ValueError(f"Source '{uid}' not found in table '{kw.SOURCES_RAYLEIGH_LGS_TABLE}' even"
                                     f" though it is of type '{t}'")

                other_data = _convert_row(hdus[kw.SOURCES_RAYLEIGH_LGS_TABLE].columns, f, kw.SOURCE_RAYLEIGH_LGS_FIELDS)
                src = aotpy.RayleighLaserGuideStar(
                    uid=uid,
                    distance=other_data[kw.SOURCE_RAYLEIGH_LGS_DISTANCE],
                    depth=other_data[kw.SOURCE_RAYLEIGH_LGS_DEPTH],
                    laser_launch_telescope=self._handle_reference(other_data[kw.LASER_LAUNCH_TELESCOPE_REFERENCE],
                                                                  kw.TELESCOPES_TABLE)
                )
            else:
                warnings.warn(f"Skipped source '{uid}': unknown type '{t}'.")
                continue

            src.right_ascension = data[kw.SOURCE_RIGHT_ASCENSION]
            src.declination = data[kw.SOURCE_DECLINATION]
            src.elevation_offset = data[kw.SOURCE_ELEVATION_OFFSET]
            src.azimuth_offset = data[kw.SOURCE_AZIMUTH_OFFSET]
            src.width = data[kw.SOURCE_WIDTH]

            self._sources[uid] = [src, True]  # sources are always referenced by AOSystem
            self._system.sources.append(src)

    def _handle_detectors(self, hdus: fits.HDUList):
        self._check_bintable(hdus, kw.DETECTORS_TABLE)

        table = hdus[kw.DETECTORS_TABLE]
        columns = table.columns
        for row in table.data:
            data = _convert_row(columns, row, kw.DETECTOR_FIELDS)

            self._detectors[data[kw.REFERENCE_UID]] = [aotpy.Detector(
                uid=data[kw.REFERENCE_UID],
                type=data[kw.DETECTOR_TYPE],
                sampling_technique=data[kw.DETECTOR_SAMPLING_TECHNIQUE],
                shutter_type=data[kw.DETECTOR_SHUTTER_TYPE],
                flat_field=self._handle_image(data[kw.DETECTOR_FLAT_FIELD]),
                readout_noise=data[kw.DETECTOR_READOUT_NOISE],
                pixel_intensities=self._handle_image(data[kw.DETECTOR_PIXEL_INTENSITIES]),
                integration_time=data[kw.DETECTOR_INTEGRATION_TIME],
                coadds=data[kw.DETECTOR_COADDS],
                dark=self._handle_image(data[kw.DETECTOR_DARK]),
                weight_map=self._handle_image(data[kw.DETECTOR_WEIGHT_MAP]),
                quantum_efficiency=data[kw.DETECTOR_QUANTUM_EFFICIENCY],
                pixel_scale=data[kw.DETECTOR_PIXEL_SCALE],
                binning=data[kw.DETECTOR_BINNING],
                bandwidth=data[kw.DETECTOR_BANDWIDTH],
                transmission_wavelength=data[kw.DETECTOR_TRANSMISSION_WAVELENGTH],
                transmission=data[kw.DETECTOR_TRANSMISSION],
                sky_background=self._handle_image(data[kw.DETECTOR_SKY_BACKGROUND]),
                gain=data[kw.DETECTOR_GAIN],
                excess_noise=data[kw.DETECTOR_EXCESS_NOISE],
                filter=data[kw.DETECTOR_FILTER],
                bad_pixel_map=self._handle_image(data[kw.DETECTOR_BAD_PIXEL_MAP]),
                dynamic_range=data[kw.DETECTOR_DYNAMIC_RANGE],
                readout_rate=data[kw.DETECTOR_READOUT_RATE],
                frame_rate=data[kw.DETECTOR_FRAME_RATE],
                transformation_matrix=self._handle_image(data[kw.TRANSFORMATION_MATRIX])
            ), False]

    def _handle_scoring_cameras(self, hdus: fits.HDUList):
        self._check_bintable(hdus, kw.SCORING_CAMERAS_TABLE)

        table = hdus[kw.SCORING_CAMERAS_TABLE]
        columns = table.columns
        for row in table.data:
            data = _convert_row(columns, row, kw.SCORING_CAMERA_FIELDS)

            self._system.scoring_cameras.append(aotpy.ScoringCamera(
                uid=data[kw.REFERENCE_UID],
                pupil_mask=self._handle_image(data[kw.SCORING_CAMERA_PUPIL_MASK]),
                wavelength=data[kw.SCORING_CAMERA_WAVELENGTH],
                transformation_matrix=self._handle_image(data[kw.TRANSFORMATION_MATRIX]),
                detector=self._handle_reference(data[kw.DETECTOR_REFERENCE], kw.DETECTORS_TABLE),
                aberration=self._handle_reference(data[kw.ABERRATION_REFERENCE], kw.ABERRATIONS_TABLE),
            ))

    def _handle_wavefront_sensors(self, hdus: fits.HDUList):
        self._check_bintable(hdus, kw.WAVEFRONT_SENSORS_TABLE)

        table = hdus[kw.WAVEFRONT_SENSORS_TABLE]
        existing_types = table.data['TYPE']
        if kw.WAVEFRONT_SENSOR_TYPE_SHACK_HARTMANN in existing_types:
            if kw.WAVEFRONT_SENSORS_SHACK_HARTMANN_TABLE in hdus:
                self._check_bintable(hdus, kw.WAVEFRONT_SENSORS_SHACK_HARTMANN_TABLE)
            else:
                raise ValueError(f"Missing table '{kw.WAVEFRONT_SENSORS_SHACK_HARTMANN_TABLE}' must exist when "
                                 f"'{kw.WAVEFRONT_SENSOR_TYPE_SHACK_HARTMANN}' type wavefront sensors exist")
        if kw.WAVEFRONT_SENSOR_TYPE_PYRAMID in existing_types:
            if kw.WAVEFRONT_SENSORS_PYRAMID_TABLE in hdus:
                self._check_bintable(hdus, kw.WAVEFRONT_SENSORS_PYRAMID_TABLE)
            else:
                raise ValueError(f"Missing table '{kw.WAVEFRONT_SENSORS_PYRAMID_TABLE}' must exist when "
                                 f"'{kw.WAVEFRONT_SENSOR_TYPE_PYRAMID}' type wavefront sensors exist")

        columns = table.columns
        for row in table.data:
            data = _convert_row(columns, row, kw.WAVEFRONT_SENSOR_FIELDS)

            uid = data[kw.REFERENCE_UID]
            t = data[kw.WAVEFRONT_SENSOR_TYPE]
            source = self._handle_reference(data[kw.SOURCE_REFERENCE], kw.SOURCES_TABLE)
            dimensions = data[kw.WAVEFRONT_SENSOR_DIMENSIONS]
            n_valid_subapertures = data[kw.WAVEFRONT_SENSOR_N_VALID_SUBAPERTURES]
            if t == kw.WAVEFRONT_SENSOR_TYPE_SHACK_HARTMANN:
                # Try to find uid in secondary table
                f = next(
                    (x for x in hdus[kw.WAVEFRONT_SENSORS_SHACK_HARTMANN_TABLE].data if x[kw.REFERENCE_UID] == uid),
                    None)
                if f is None:
                    raise ValueError(f"Wavefront sensor '{uid}' not found in table "
                                     f"'{kw.WAVEFRONT_SENSORS_SHACK_HARTMANN_TABLE}' even though it is of type '{t}'")

                other_data = _convert_row(hdus[kw.WAVEFRONT_SENSORS_SHACK_HARTMANN_TABLE].columns, f,
                                          kw.WAVEFRONT_SENSOR_SHACK_HARTMANN_FIELDS)
                if dimensions != 2:
                    warnings.warn(f"Unexpected value for '{kw.WAVEFRONT_SENSOR_DIMENSIONS}' in wavefront sensor '{uid}'"
                                  f" of type '{t}'. Expected 2, got {dimensions}.")
                wfs = aotpy.ShackHartmann(
                    uid=uid,
                    source=source,
                    n_valid_subapertures=n_valid_subapertures,
                    centroiding_algorithm=other_data[kw.WAVEFRONT_SENSOR_SHACK_HARTMANN_CENTROIDING_ALGORITHM],
                    centroid_gains=self._handle_image(other_data[kw.WAVEFRONT_SENSOR_SHACK_HARTMANN_CENTROID_GAINS]),
                    spot_fwhm=self._handle_image(other_data[kw.WAVEFRONT_SENSOR_SHACK_HARTMANN_SPOT_FWHM])
                )
            elif t == kw.WAVEFRONT_SENSOR_TYPE_PYRAMID:
                # Try to find uid in secondary table
                f = next((x for x in hdus[kw.WAVEFRONT_SENSORS_PYRAMID_TABLE].data if x[kw.REFERENCE_UID] == uid), None)
                if f is None:
                    raise ValueError(f"Wavefront sensor '{uid}' not found in table "
                                     f"'{kw.WAVEFRONT_SENSORS_PYRAMID_TABLE}' even though it is of type '{t}'")

                other_data = _convert_row(hdus[kw.WAVEFRONT_SENSORS_PYRAMID_TABLE].columns, f,
                                          kw.WAVEFRONT_SENSOR_PYRAMID_FIELDS)
                wfs = aotpy.Pyramid(
                    uid=uid,
                    source=source,
                    dimensions=dimensions,
                    n_valid_subapertures=n_valid_subapertures,
                    n_sides=other_data[kw.WAVEFRONT_SENSOR_PYRAMID_N_SIDES],
                    modulation=other_data[kw.WAVEFRONT_SENSOR_PYRAMID_MODULATION]
                )
            else:
                warnings.warn(f"Skipped wavefront sensor '{uid}': unknown type '{t}'.")
                continue

            wfs.measurements = self._handle_image(data[kw.WAVEFRONT_SENSOR_MEASUREMENTS])
            wfs.ref_measurements = self._handle_image(data[kw.WAVEFRONT_SENSOR_REF_MEASUREMENTS])
            wfs.subaperture_mask = self._handle_image(data[kw.WAVEFRONT_SENSOR_SUBAPERTURE_MASK])
            wfs.mask_offsets = [aotpy.Coordinates(x, y) for x, y in zip(data[kw.WAVEFRONT_SENSOR_MASK_X_OFFSETS],
                                                                        data[kw.WAVEFRONT_SENSOR_MASK_Y_OFFSETS])]
            wfs.subaperture_size = data[kw.WAVEFRONT_SENSOR_SUBAPERTURE_SIZE]
            wfs.subaperture_intensities = self._handle_image(data[kw.WAVEFRONT_SENSOR_SUBAPERTURE_INTENSITIES])
            wfs.wavelength = data[kw.WAVEFRONT_SENSOR_WAVELENGTH]
            wfs.optical_gain = self._handle_image(data[kw.WAVEFRONT_SENSOR_OPTICAL_GAIN])
            wfs.transformation_matrix = self._handle_image(data[kw.TRANSFORMATION_MATRIX])
            wfs.detector = self._handle_reference(data[kw.DETECTOR_REFERENCE], kw.DETECTORS_TABLE)
            wfs.aberration = self._handle_reference(data[kw.ABERRATION_REFERENCE], kw.ABERRATIONS_TABLE)
            wfs.non_common_path_aberration = self._handle_reference(data[kw.NCPA_REFERENCE], kw.ABERRATIONS_TABLE)

            self._wfss[uid] = [wfs, True]  # wavefront sensors are always referenced by AOSystem
            self._system.wavefront_sensors.append(wfs)

    def _handle_wavefront_correctors(self, hdus: fits.HDUList):
        self._check_bintable(hdus, kw.WAVEFRONT_CORRECTORS_TABLE)

        table = hdus[kw.WAVEFRONT_CORRECTORS_TABLE]
        existing_types = table.data['TYPE']
        if kw.WAVEFRONT_CORRECTOR_TYPE_DM in existing_types:
            if kw.WAVEFRONT_CORRECTORS_DM_TABLE in hdus:
                self._check_bintable(hdus, kw.WAVEFRONT_CORRECTORS_DM_TABLE)
            else:
                raise ValueError(f"Missing table '{kw.WAVEFRONT_CORRECTORS_DM_TABLE}' must exist when "
                                 f"'{kw.WAVEFRONT_CORRECTOR_TYPE_DM}' type wavefront correctors exist")

        columns = table.columns
        for row in table.data:
            data = _convert_row(columns, row, kw.WAVEFRONT_CORRECTOR_FIELDS)

            uid = data[kw.REFERENCE_UID]
            t = data[kw.WAVEFRONT_CORRECTOR_TYPE]
            telescope = self._handle_reference(data[kw.TELESCOPE_REFERENCE], kw.TELESCOPES_TABLE)

            if t == kw.WAVEFRONT_CORRECTOR_TYPE_DM:
                # Try to find uid in secondary table
                f = next((x for x in hdus[kw.WAVEFRONT_CORRECTORS_DM_TABLE].data if x[kw.REFERENCE_UID] == uid), None)
                if f is None:
                    raise ValueError(f"Wavefront corrector '{uid}' not found in table "
                                     f"'{kw.WAVEFRONT_CORRECTORS_DM_TABLE}' even though it is of type '{t}'")

                other_data = _convert_row(hdus[kw.WAVEFRONT_CORRECTORS_DM_TABLE].columns, f,
                                          kw.WAVEFRONT_CORRECTOR_DM_FIELDS)
                cor = aotpy.DeformableMirror(
                    uid=uid,
                    telescope=telescope,
                    n_valid_actuators=data[kw.WAVEFRONT_CORRECTOR_N_VALID_ACTUATORS],
                    actuator_coordinates=[aotpy.Coordinates(x, y) for x, y in
                                          zip(other_data[kw.WAVEFRONT_CORRECTOR_DM_ACTUATORS_X],
                                              other_data[kw.WAVEFRONT_CORRECTOR_DM_ACTUATORS_Y])],
                    influence_function=self._handle_image(other_data[kw.WAVEFRONT_CORRECTOR_DM_INFLUENCE_FUNCTION]),
                    stroke=other_data[kw.WAVEFRONT_CORRECTOR_DM_STROKE]
                )
            elif t == kw.WAVEFRONT_CORRECTOR_TYPE_TTM:
                cor = aotpy.TipTiltMirror(
                    uid=uid,
                    telescope=telescope
                )
            elif t == kw.WAVEFRONT_CORRECTOR_TYPE_LS:
                cor = aotpy.LinearStage(
                    uid=uid,
                    telescope=telescope
                )
            else:
                warnings.warn(f"Skipped wavefront corrector '{uid}': unknown type '{t}'.")
                continue

            cor.pupil_mask = self._handle_image(data[kw.WAVEFRONT_CORRECTOR_PUPIL_MASK])
            cor.tfz_num = data[kw.WAVEFRONT_CORRECTOR_TFZ_NUM]
            cor.tfz_den = data[kw.WAVEFRONT_CORRECTOR_TFZ_DEN]
            cor.transformation_matrix = self._handle_image(data[kw.TRANSFORMATION_MATRIX])
            cor.aberration = self._handle_reference(data[kw.ABERRATION_REFERENCE], kw.ABERRATIONS_TABLE)

            self._wfcs[uid] = [cor, True]  # wavefront corectors are always referenced by AOSystem
            self._system.wavefront_correctors.append(cor)

    def _handle_loops(self, hdus: fits.HDUList):
        self._check_bintable(hdus, kw.LOOPS_TABLE)

        table = hdus[kw.LOOPS_TABLE]
        existing_types = table.data['TYPE']
        if kw.LOOPS_TYPE_CONTROL in existing_types:
            if kw.LOOPS_CONTROL_TABLE in hdus:
                self._check_bintable(hdus, kw.LOOPS_CONTROL_TABLE)
            else:
                raise ValueError(f"Missing table '{kw.LOOPS_CONTROL_TABLE}' must exist when "
                                 f"'{kw.LOOPS_TYPE_CONTROL}' type loops exist")
        if kw.LOOPS_TYPE_OFFLOAD in existing_types:
            if kw.LOOPS_OFFLOAD_TABLE in hdus:
                self._check_bintable(hdus, kw.LOOPS_OFFLOAD_TABLE)
            else:
                raise ValueError(f"Missing table '{kw.LOOPS_OFFLOAD_TABLE}' must exist when "
                                 f"'{kw.LOOPS_TYPE_OFFLOAD}' type loops exist")

        columns = table.columns
        for row in table.data:
            data = _convert_row(columns, row, kw.LOOPS_FIELDS)

            uid = data[kw.REFERENCE_UID]
            t = data[kw.WAVEFRONT_SENSOR_TYPE]
            commanded = self._handle_reference(data[kw.LOOPS_COMMANDED], kw.WAVEFRONT_CORRECTORS_TABLE)

            if t == kw.LOOPS_TYPE_CONTROL:
                # Try to find uid in secondary table
                f = next((x for x in hdus[kw.LOOPS_CONTROL_TABLE].data if x[kw.REFERENCE_UID] == uid), None)
                if f is None:
                    raise ValueError(f"Loop '{uid}' not found in table "
                                     f"'{kw.LOOPS_CONTROL_TABLE}' even though it is of type '{t}'")

                other_data = _convert_row(hdus[kw.LOOPS_CONTROL_TABLE].columns, f, kw.LOOPS_CONTROL_FIELDS)
                loop = aotpy.ControlLoop(
                    uid=uid,
                    commanded_corrector=commanded,
                    input_sensor=self._handle_reference(other_data[kw.LOOPS_CONTROL_INPUT_SENSOR],
                                                        kw.WAVEFRONT_SENSORS_TABLE),
                    modes=self._handle_image(other_data[kw.LOOPS_CONTROL_MODES]),
                    modal_coefficients=self._handle_image(other_data[kw.LOOPS_CONTROL_MODAL_COEFFICIENTS]),
                    control_matrix=self._handle_image(other_data[kw.LOOPS_CONTROL_CONTROL_MATRIX]),
                    measurements_to_modes=self._handle_image(other_data[kw.LOOPS_CONTROL_MEASUREMENTS_TO_MODES]),
                    modes_to_commands=self._handle_image(other_data[kw.LOOPS_CONTROL_MODES_TO_COMMANDS]),
                    interaction_matrix=self._handle_image(other_data[kw.LOOPS_CONTROL_INTERACTION_MATRIX]),
                    commands_to_modes=self._handle_image(other_data[kw.LOOPS_CONTROL_COMMANDS_TO_MODES]),
                    modes_to_measurements=self._handle_image(other_data[kw.LOOPS_CONTROL_MODES_TO_MEASUREMENTS]),
                    residual_commands=self._handle_image(other_data[kw.LOOPS_CONTROL_RESIDUAL_COMMANDS])
                )
            elif t == kw.LOOPS_TYPE_OFFLOAD:
                # Try to find uid in secondary table
                f = next((x for x in hdus[kw.LOOPS_OFFLOAD_TABLE].data if x[kw.REFERENCE_UID] == uid), None)
                if f is None:
                    raise ValueError(f"Loop '{uid}' not found in table "
                                     f"'{kw.LOOPS_OFFLOAD_TABLE}' even though it is of type '{t}'")

                other_data = _convert_row(hdus[kw.LOOPS_OFFLOAD_TABLE].columns, f, kw.LOOPS_OFFLOAD_FIELDS)
                loop = aotpy.OffloadLoop(
                    uid=uid,
                    commanded_corrector=commanded,
                    input_corrector=self._handle_reference(other_data[kw.LOOPS_OFFLOAD_INPUT_CORRECTOR],
                                                           kw.WAVEFRONT_CORRECTORS_TABLE),
                    offload_matrix=self._handle_image(other_data[kw.LOOPS_OFFLOAD_OFFLOAD_MATRIX])
                )
            else:
                warnings.warn(f"Skipped loop '{uid}': unknown type '{t}'.")
                continue

            loop.time = self._handle_reference(data[kw.TIME_REFERENCE], kw.TIME_TABLE)
            if (status := data[kw.LOOPS_STATUS]) is None:
                loop.closed = None
            elif status == kw.LOOPS_STATUS_CLOSED:
                loop.closed = True
            elif status == kw.LOOPS_STATUS_OPEN:
                loop.closed = False
            else:
                warnings.warn(f"Ignored unknown loop status '{status}'.")
                loop.closed = None
            loop.commands = self._handle_image(data[kw.LOOPS_COMMANDS])
            loop.ref_commands = self._handle_image(data[kw.LOOPS_REF_COMMANDS])
            loop.framerate = data[kw.LOOPS_FRAMERATE]
            loop.delay = data[kw.LOOPS_DELAY]
            loop.time_filter_num = self._handle_image(data[kw.LOOPS_TIME_FILTER_NUM])
            loop.time_filter_den = self._handle_image(data[kw.LOOPS_TIME_FILTER_DEN])

            self._system.loops.append(loop)



"""
This module contains a class for translating data produced by ESO's GALACSI system.
"""

import importlib.resources
import warnings
from datetime import datetime
from pathlib import Path

import numpy as np
from astropy.io import fits

import aotpy
from aotpy.io import image_from_file
from .eso import ESOTranslator


# TODO set image units


class GALACSITranslator(ESOTranslator):
    """Contains functions for translating telemetry data produced by ESO's GALACSI system.

    Parameters
    ----------
    path_lgs
        Path to folder containing LGS data (LGSAcq, LGSCtr, LGSRecn, JitCtr, JitRecnOptimiser, RTC).
    path_ir
        Path to folder containing IR data (IRAcq, IRCtr, IRLoopMonitor).
    path_pix
        Path to folder containing pixel data.
    """

    def __init__(self, path_lgs: str, path_ir: str, path_pix: str):
        self.system = aotpy.AOSystem(ao_mode='LTAO', name='GALACSI')
        self.system.main_telescope = aotpy.MainTelescope(
            uid='ESO VLT UT4',
            enclosing_diameter=8.2,
            inscribed_diameter=8.2
        )

        self._handle_lgs_data(path_lgs)
        self._handle_ngs_data(path_ir, path_pix)

    def _handle_lgs_data(self, path_lgs):
        path_lgs = Path(path_lgs)
        lgs_loop_frame = fits.getdata(path_lgs / f'{path_lgs.name}.fits', extname='LGSLoopFrame')

        self.dsm_valid = fits.getdata(path_lgs / 'RTC.USED_ACT_MAP.fits')[0] - 1
        # We have to subtract one because the array uses one-based indexing unlike Python

        self.dsm = aotpy.DeformableMirror(
            uid='DSM',
            telescope=self.system.main_telescope,
            n_valid_actuators=self.dsm_valid.size,
        )
        self.system.wavefront_correctors.append(self.dsm)

        lgs_timestamps = lgs_loop_frame['Seconds'] + lgs_loop_frame['USeconds'] / 1.e6
        self.system.date_beginning = datetime.utcfromtimestamp(lgs_timestamps[0])
        self.system.date_end = datetime.utcfromtimestamp(lgs_timestamps[-1])
        lgs_frame_numbers = lgs_loop_frame['FrameCounter']
        lgs_time = aotpy.Time('LGS Loop Time', timestamps=lgs_timestamps.tolist(),
                              frame_numbers=lgs_frame_numbers.tolist())

        aof_data_path = importlib.resources.files('aotpy.data') / 'GALACSI'
        with importlib.resources.as_file(aof_data_path / 'subap.fits') as p:
            subaperture_mask = image_from_file(p)
        n_valid_subapertures = np.count_nonzero(subaperture_mask.data != -1)

        dsm_positions = aotpy.Image('DSM_positions', lgs_loop_frame['DSM_Positions'][:, self.dsm_valid])
        m2c = image_from_file(path_lgs / 'LGSCtr.ACT_POS_MODAL_PROJECTION.fits')
        lgs_tfz_num = aotpy.Image('LGSCtr.A_TERMS', fits.getdata(path_lgs / 'LGSCtr.A_TERMS.fits').T)
        lgs_tfz_den = aotpy.Image('LGSCtr.B_TERMS', fits.getdata(path_lgs / 'LGSCtr.B_TERMS.fits').T)
        jit_tfz_num = fits.getdata(path_lgs / 'JitCtr.A_TERMS.fits').T
        jit_tfz_den = fits.getdata(path_lgs / 'JitCtr.B_TERMS.fits').T
        jit_ref = fits.getdata(path_lgs / 'JitCtr.ACT_POS_REF_MAP_WITH_OFFSETS.fits')[:, 0]
        off_ref = fits.getdata(path_lgs / 'JitCtr.OACT_POS_REF_MAP.fits')[0, :]
        proj_map = fits.getdata(path_lgs / f'JitCtr.PROJ_MAP_SCALED.fits')
        im_list = np.split(fits.getdata(path_lgs / f'RTC.IMref4Atm.fits')[:, self.dsm_valid], 4, axis=0)
        for i in range(1, 5):
            llt = aotpy.LaserLaunchTelescope(f'LLT{i}')
            lgs = aotpy.SodiumLaserGuideStar(uid=f'LGS{i}', laser_launch_telescope=llt)
            self.system.sources.append(lgs)

            gradients = self._stack_slopes(lgs_loop_frame[f'WFS{i}_Gradients'], slope_axis=1)
            reference = self._stack_slopes(fits.getdata(path_lgs / f'LGSAcq.DET{i}.REFSLP_WITH_OFFSETS.fits'),
                                           slope_axis=1)[0]
            wfs = aotpy.ShackHartmann(
                uid=f'LGS WFS{i}',
                source=lgs,
                n_valid_subapertures=n_valid_subapertures,
                measurements=aotpy.Image(f'WFS{i}_Gradients', gradients),
                ref_measurements=aotpy.Image(f'LGSAcq.DET{i}.REFSLP_WITH_OFFSETS', reference),
                subaperture_mask=subaperture_mask,
                subaperture_intensities=aotpy.Image(f'WFS{i}_Intensities', lgs_loop_frame[f'WFS{i}_Intensities'])
            )

            wfs.detector = aotpy.Detector(
                uid=f'LGS DET{i}',
                dark=image_from_file(path_lgs / f'LGSAcq.DET{i}.DARK.fits'),
                weight_map=image_from_file(path_lgs / f'LGSAcq.DET{i}.WEIGHT.fits')
            )
            self.system.wavefront_sensors.append(wfs)

            cm = self._stack_slopes(fits.getdata(path_lgs / f'LGSRecn.REC{i}.HOCM.fits')[self.dsm_valid], slope_axis=1)
            im = self._stack_slopes(im_list[i - 1], slope_axis=0)
            self.system.loops.append(aotpy.ControlLoop(
                uid=f'High-order loop {i}',
                input_sensor=wfs,
                commanded_corrector=self.dsm,
                commands=dsm_positions,
                time=lgs_time,
                framerate=1000,
                time_filter_num=lgs_tfz_num,
                time_filter_den=lgs_tfz_den,
                control_matrix=aotpy.Image(f'LGSRecn.REC{i}.HOCM', cm),
                interaction_matrix=aotpy.Image(f'Interaction_Matrix_LGS_WFS{i}', im)
            ))

            jit = aotpy.TipTiltMirror(
                uid=f'Jitter{i}',
                telescope=llt
            )

            cm = self._stack_slopes(fits.getdata(path_lgs / f'JitRecnOptimiser.JitCM{i}.fits'), slope_axis=1)
            self.system.loops.append(aotpy.ControlLoop(
                uid=f'Jitter loop {i}',
                input_sensor=wfs,
                commanded_corrector=jit,
                commands=aotpy.Image(f'Jitter{i}_Positions', lgs_loop_frame[f'Jitter{i}_Positions']),
                ref_commands=aotpy.Image(f'Jit{i}Ctr.ACT_POS_REF_MAP_WITH_OFFSETS', jit_ref[(i - 1) * 2: i * 2]),
                time=lgs_time,
                framerate=1000,
                time_filter_num=aotpy.Image(f'Jit{i}Ctr.A_TERMS', jit_tfz_num[i - 1:i + 1, :]),
                time_filter_den=aotpy.Image(f'Jit{i}Ctr.B_TERMS', jit_tfz_den[i - 1:i + 1, :]),
                control_matrix=aotpy.Image(f'JitRecnOptimiser.JitCM{i}', cm),
                modes_to_commands=m2c
            ))

            fsm = aotpy.TipTiltMirror(
                uid=f'Field Steering Mirror {i}',
                telescope=llt
            )

            self.system.loops.append(aotpy.OffloadLoop(
                uid=f'Jitter Offload loop {i}',
                input_corrector=jit,
                commanded_corrector=fsm,
                commands=aotpy.Image(f'Jitter{i}_Offload', lgs_loop_frame[f'Jitter{i}_Offload']),
                ref_commands=aotpy.Image(f'Jit{i}Ctr.OACT_POS_REF_MAP', off_ref[(i - 1) * 2: i * 2]),
                time=lgs_time,
                offload_matrix=aotpy.Image(f'Jitter{i}_Offload_Matrix', proj_map[(i - 1) * 2:i * 2, (i - 1) * 2:i * 2])
            ))

            self.system.wavefront_correctors.extend([jit, fsm])

    def _handle_ngs_data(self, path_ir, path_pix):
        path_ir = Path(path_ir)
        ir_loop_frame = fits.getdata(path_ir / f'{path_ir.name}.fits', extname='IRLoopFrame')
        path_pix = Path(path_pix)
        pix_loop_frame = fits.getdata(path_pix / f'{path_pix.name}.fits', extname='IRPixelFrame')

        ngs_timestamps = ir_loop_frame['Seconds'] + ir_loop_frame['USeconds'] / 1.e6
        if np.all(ngs_timestamps == 0):
            # The file has no timestamps
            ngs_timestamps_list = []
        else:
            ngs_timestamps_list = ngs_timestamps.tolist()
        ho_frame_numbers = ir_loop_frame['HOFrameCounter']
        ir_time = aotpy.Time('NGS Loop Time', timestamps=ngs_timestamps_list, frame_numbers=ho_frame_numbers.tolist())

        ngs = aotpy.NaturalGuideStar('NGS')
        self.system.sources.append(ngs)

        gradients = self._stack_slopes(ir_loop_frame['WFS_Gradients'], slope_axis=1)
        reference = self._stack_slopes(fits.getdata(path_ir / 'IRAcq.DET1.REFSLP_WITH_OFFSETS.fits'), slope_axis=1)[0]
        ngs_wfs = aotpy.ShackHartmann(
            uid='NGS WFS1',
            n_valid_subapertures=4,  # All subapertures are valid
            subaperture_mask=aotpy.Image('NGS_WFS_SUBAPERTURE_MASK', np.array([[1, 3], [2, 4]])),
            source=ngs,
            measurements=aotpy.Image('NGS_WFS_Gradients', gradients),
            ref_measurements=aotpy.Image('IRAcq.DET1.REFSLP_WITH_OFFSETS', reference),
            subaperture_intensities=aotpy.Image('WFS_Intensities', ir_loop_frame['WFS_Intensities'])
        )
        self.system.wavefront_sensors.append(ngs_wfs)

        # Find the indexes where the counter for pixels matches the counter for ngs
        # Assumes the frames for pixel are contained in the frames for ngs
        ir_fc = ir_loop_frame['FrameCounter']
        pix_fc = pix_loop_frame['FrameCounter']
        if pix_fc[0] < ir_fc[0] or pix_fc[-1] > ir_fc[-1]:
            warnings.warn('Pixel frame counter not contained in IR frame counter, pixel time data may be incorrect.')
        pix_time_mask = np.searchsorted(ir_fc, pix_fc)
        pix_timestamps = ngs_timestamps[pix_time_mask]
        if np.all(pix_timestamps == 0):
            # The file has no timestamps
            pix_timestamps_list = []
        else:
            pix_timestamps_list = pix_timestamps.tolist()

        pix_time = aotpy.Time('Pixel Time', timestamps=pix_timestamps_list,
                              frame_numbers=ho_frame_numbers[pix_time_mask].tolist())

        ngs_wfs.detector = aotpy.Detector(
            uid='NGS DET1',
            dark=image_from_file(path_ir / 'IRAcq.DET1.DARK.fits'),
            weight_map=image_from_file(path_ir / 'IRAcq.DET1.WEIGHT.fits'),
            pixel_intensities=aotpy.Image(name='NGS Pixels',
                                          data=self._get_pixel_data_from_table(pix_loop_frame),
                                          time=pix_time)
        )

        s2m = self._stack_slopes(fits.getdata(path_ir / 'IRCtr.SENSOR_2_MODES.fits'), slope_axis=1)
        m2c = fits.getdata(path_ir / 'IRCtr.MODES_2_ACT.fits')[self.dsm_valid]
        self.system.loops.append(aotpy.ControlLoop(
            uid='Low-order loop',
            input_sensor=ngs_wfs,
            commanded_corrector=self.dsm,
            commands=aotpy.Image('LO_Positions', ir_loop_frame['LO_Positions'][:, self.dsm_valid]),
            measurements_to_modes=aotpy.Image('IRCtr.SENSOR_2_MODES', s2m),
            modes_to_commands=aotpy.Image('IRCtr.MODES_2_ACT', m2c),
            time=ir_time,
            time_filter_num=aotpy.Image('IRCtr.A_TERMS', fits.getdata(path_ir / 'IRCtr.A_TERMS.fits').T),
            time_filter_den=aotpy.Image('IRCtr.B_TERMS', fits.getdata(path_ir / 'IRCtr.B_TERMS.fits').T)
        ))

    def _get_eso_telescope_name(self) -> str:
        return 'ESO-VLT-U4'

    def _get_eso_ao_name(self) -> str:
        return 'GALACSI'

    def _get_run_id(self) -> str:
        return '60.A-9278(B)'



"""
This module contains base classes for defining writers and readers in aotpy.
"""

import os
from abc import ABC, abstractmethod

import aotpy


class SystemWriter(ABC):
    """Abstract class for system writers.

    Writers are able to write an `AOSystem` into a file using some format.

    Parameters
    ----------
    system
        `AOSystem` to be written into a file.

    """

    @abstractmethod
    def __init__(self, system: aotpy.AOSystem) -> None:
        self._system = system

    @abstractmethod
    def write(self, filename: str | os.PathLike, **kwargs) -> None:
        """
        Write the initialized `system` into the specified `filename`.

        Parameters
        ----------
        filename
            Path to the file that will be written.
        **kwargs
            Keyword arguments passed on as options to the file handling function.
        """
        pass


class SystemReader(ABC):
    """Abstract class for system readers.

    Readers are able to create an `AOSystem` based on a file of a certain format.

    Parameters
    ----------
    filename
        Path to file to be read into an `AOSystem`.
    extra_data : default = False
        Whether it is expected that the file contains some data that does not fit the AOT standard. If `extra_data` is
        not `True`, user will be warned if extra data is detected.
    **kwargs
        Keyword arguments passed on as options to the file handling function.
    """

    def __init__(self, filename: str | os.PathLike, *, extra_data: bool = False, **kwargs) -> None:
        self._filename = filename
        self._extra_data_flag = extra_data

        self._initialize_data()
        self._system, self._extra_data = self._read(**kwargs)

    def get_system(self) -> aotpy.AOSystem:
        """
        Return `AOSystem` that has been read.
        """
        return self._system

    def get_extra_data(self) -> list | None:
        """
        Return a list of extra data that may have been in file. If no extra data exists, `None` is returned.
        """
        return self._extra_data

    @abstractmethod
    def _initialize_data(self) -> None:
        """
        Initialize data structures necessary for reading the file.
        """
        pass

    @abstractmethod
    def _read(self, **kwargs) -> tuple[aotpy.AOSystem, list]:
        """
        Read file and build `AOSystem` that contains the data in the initialized `filename`.
        Return the system that was built along with any extra data that may be present.

        Parameters
        ----------
        **kwargs
            Keyword arguments passed on as options to the file handling function.
        """
        pass



"""
This subpackage contains modules with translators, which are able to convert non-standard AO telemetry files into aotpy
objects.

The aotpy objects created by these translators can be handled as any other aotpy object regardless of origin.
"""

from .galacsi import GALACSITranslator
from .ciao import CIAOTranslator
from .eris import ERISTranslator
from .naomi import NAOMITranslator
from .papyrus import PAPYRUSTranslator



"""
This module contains classes that describe different types of wavefront correctors in a system.
"""

from dataclasses import dataclass, field

from .aberration import Aberration
from .base import Referenceable, Coordinates
from .image import Image
from .telescope import Telescope

__all__ = ['WavefrontCorrector', 'DeformableMirror', 'TipTiltMirror', 'LinearStage']


@dataclass(kw_only=True)
class WavefrontCorrector(Referenceable):
    """Abstract class that contains data related to a wavefront corrector in the system."""
    telescope: Telescope
    n_valid_actuators: int
    """Number of valid actuators in the corrector. (in count units)"""

    pupil_mask: Image = None
    """Binary image that describes the shape of the pupil. A 1 indicates the presence of the pupil, while a 0 indicates
    the opposite. (Dimensions :math:`h \\times w`, dimensionless quantity, using data type int)"""

    tfz_num: list[float] = field(default_factory=list)
    'List of numerators of the transfer function Z. (dimensionless quantity)'

    tfz_den: list[float] = field(default_factory=list)
    'List of denominators of the transfer function Z. (dimensionless quantity)'

    transformation_matrix: Image = None
    r"""Matrix that defines 2-dimensional affine transformations over time (:math:`t`) using homogeneous coordinates.
    Any combination of translation, reflection, scale, rotation and shearing can be described via a single
    :math:`3 \times 3` matrix :math:`M` such that :math:`P' = MP`, where :math:`P` is a
    :math:`\begin{bmatrix}x & y & 1 \end{bmatrix}` vector (with :math:`x` and :math:`y` being the original horizontal 
    and vertical coordinates, respectively) and :math:`P'` is a :math:`\begin{bmatrix}x' & y' & 1 \end{bmatrix}`, where
    :math:`x'` and :math:`y'` are the transformed coordinates. All geometry information must be described relative to
    the same reference origin point, from which transformations may occur.
    (Dimensions :math:`3 \times 3 \times t`, dimensionless quantity, using data type flt)"""

    aberration: Aberration = None

    def __post_init__(self):
        if self.__class__ == WavefrontCorrector:
            raise TypeError("Cannot instantiate abstract class.")


@dataclass(kw_only=True)
class DeformableMirror(WavefrontCorrector):
    """Contains data related to a deformable mirror in the system."""
    actuator_coordinates: list[Coordinates] = field(default_factory=list)
    'List of horizontal/vertical coordinates of the valid actuators of the DM. (in m units)'

    influence_function: Image = None
    """A set of 2D images, one for each valid actuator, where each image represents the displacement of the surface of
    the deformable mirror after poking the respective actuator.
    (Dimensions :math:`a_v \\times h \\times w`, in m units, using data type flt)"""

    stroke: float = None
    'Maximum possible actuator displacement, measured as an excursion from a central null position. (in m units)'


@dataclass(kw_only=True)
class TipTiltMirror(WavefrontCorrector):
    """Contains data related to a tip-tilt mirror in the system. It is assumed it has a two-axis movement."""
    n_valid_actuators: int = field(init=False, default=2)
    pass


@dataclass(kw_only=True)
class LinearStage(WavefrontCorrector):
    """Contains data related to a linear stage in the system. It is assumed it has a single axis of movement."""
    n_valid_actuators: int = field(init=False, default=1)
    pass



"""
This module contains a class for translating data produced by ALPAO's PAPYRUS system. It assumes the MATLAB files
produced by PAPYRUS have been converted to use structs instead of classes.
"""

import importlib.resources

import numpy as np

import aotpy
from aotpy.io import image_from_file
from .base import BaseTranslator

try:
    from scipy.io import loadmat
except (ImportError, ModuleNotFoundError):
    loadmat = None


class PAPYRUSTranslator(BaseTranslator):
    def __init__(self, file) -> None:
        if loadmat is None:
            raise ImportError("Translating PAPYRUS data requires the scipy module.")
        data = loadmat(file, simplify_cells=True)['data']

        self.system = aotpy.AOSystem(ao_mode='SCAO', name='PAPYRUS')

        self.system.sources = [aotpy.NaturalGuideStar(uid=data['source'])]

        papyrus_data_path = importlib.resources.files('aotpy.data') / 'PAPYRUS'
        with importlib.resources.as_file(papyrus_data_path / 'T152_pupil.fits') as p:
            pupil_mask = image_from_file(p, name='T152 PUPIL')

        self.system.main_telescope = aotpy.MainTelescope(
            uid="T152",
            enclosing_diameter=1.5,
            inscribed_diameter=1.5,
            pupil_mask=pupil_mask
        )

        actuator_coordinates = [aotpy.Coordinates(coor[0], coor[1]) for coor in data['wfcCommandStruct']['coordinates']]
        dm = aotpy.DeformableMirror(uid="DM_ALPAO_241",
                                    telescope=self.system.main_telescope,
                                    n_valid_actuators=data['wfc']['offsetStruct']['values'].size,
                                    actuator_coordinates=actuator_coordinates,
                                    pupil_mask=self.system.main_telescope.pupil_mask)
        self.system.wavefront_correctors.append(dm)

        slope_x = data['wfsSlopesStruct']['sx']
        slope_y = data['wfsSlopesStruct']['sy']
        measurements = np.empty((slope_x.shape[1], 2, slope_x.shape[0]))
        measurements[:, 0, :] = np.transpose(slope_x)
        measurements[:, 1, :] = np.transpose(slope_y)
        sh = aotpy.ShackHartmann(uid=data['wfsUid'],
                                 source=self.system.sources[0],
                                 n_valid_subapertures=data['wfsSlopesStruct']['mask'].sum(),
                                 measurements=aotpy.Image("papyrusShSlopes", measurements))

        detector = aotpy.Detector(uid='cblueOne',
                                  readout_noise=3,
                                  pixel_intensities=aotpy.Image("cblueFrames", np.moveaxis(data['wfsImages'], -1, 0)))
        sh.detector = detector
        self.system.wavefront_sensors.append(sh)

        im = data['calibrator']['interactionMatrix']
        interaction_matrix = np.empty((im.shape[0] // 2, 2, im.shape[1]))
        interaction_matrix[:, 0, :] = im[::2]
        interaction_matrix[:, 1, :] = im[1::2]

        cm = data['loop']['commandMatrix']
        control_matrix = np.empty((cm.shape[0], 2, cm.shape[1] // 2))
        control_matrix[:, 0, :] = cm[:, ::2]
        control_matrix[:, 1, :] = cm[:, 1::2]

        loop = aotpy.ControlLoop(
            uid=data['uid'],
            commanded_corrector=dm,
            input_sensor=sh,
            commands=aotpy.Image('papyrusDmCommands', np.transpose(data['wfcCommandStruct']['values'])),
            interaction_matrix=aotpy.Image("papyrusInteractionMatrix", interaction_matrix),
            control_matrix=aotpy.Image("papyruscontrolMatrix", control_matrix))
        self.system.loops.append(loop)



"""
This module contains a class for describing the passage of time related to data in the system.
"""

from dataclasses import dataclass, field

from .base import Referenceable

__all__ = ['Time']


@dataclass(kw_only=True)
class Time(Referenceable):
    """Contains data that describes the passage of time. All time in a system must be synchronous.
    Can be associated with time-varying data."""

    timestamps: list[float] = field(default_factory=list)
    'List of Unix timestamps at which the respective data applies. (in s units)'

    frame_numbers: list[float] = field(default_factory=list)
    'List of frame numbers at which the respective data applies. (in count units)'



"""
This module contains a class for translating data produced by ESO's NAOMI system.
"""

import importlib.resources
import warnings
from datetime import datetime
from pathlib import Path

import numpy as np
from astropy.io import fits

import aotpy
from aotpy.io import image_from_file
from .eso import ESOTranslator


# TODO set image units


class NAOMITranslator(ESOTranslator):
    """Contains functions for translating telemetry data produced by ESO's NAOMI system.

    Parameters
    ----------
    path
        Path to folder containing telemetry data.
    at_number : {1, 2, 3, 4}
        Number of the AT that produced the data.
    """

    def __init__(self, path: str, at_number: int):
        path = Path(path)
        self._at_number = at_number

        with fits.open(path / 'NAOMI_LOOP_0001.fits', extname='LoopFrame') as hdus:
            main_hdr = hdus[0].header
            main_loop_frame: fits.FITS_rec = hdus['LoopFrame'].data

        self.system = aotpy.AOSystem(ao_mode='SCAO', name='NAOMI')
        self.system.main_telescope = aotpy.MainTelescope(
            uid=f'ESO VLT AT{at_number}',
            elevation=main_hdr['ESO TEL ALT'],
            azimuth=self._azimuth_conversion(main_hdr['ESO TEL AZ']),
            parallactic=main_hdr['ESO TEL PRLTIC'],
            enclosing_diameter=1.82,
            inscribed_diameter=1.82
        )
        naomi_data_path = importlib.resources.files('aotpy.data') / 'NAOMI'
        with importlib.resources.as_file(naomi_data_path / 'zernike_control_modes.fits') as p:
            # Load file with the representation of the modes controlled in NAOMI (Zernike modes 2 to 15)
            control_modes = aotpy.Image('CONTROL MODES', fits.getdata(p))

        if main_hdr['ESO AOS CM MODES CONTROLLED'] != control_modes.data.shape[0]:
            warnings.warn("Keyword 'ESO AOS CM MODES CONTROLLED' does not match expected number of control modes.")

        ngs = aotpy.NaturalGuideStar(uid='NGS',
                                     right_ascension=main_hdr['RA'],
                                     declination=main_hdr['DEC'])

        main_timestamps = main_loop_frame['Seconds'] + main_loop_frame['USeconds'] / 1.e6
        self.system.date_beginning = datetime.utcfromtimestamp(main_timestamps[0])
        self.system.date_end = datetime.utcfromtimestamp(main_timestamps[-1])
        main_frame_numbers = main_loop_frame['FrameCounter']
        loop_time = aotpy.Time('Loop Time', timestamps=main_timestamps.tolist(),
                               frame_numbers=main_frame_numbers.tolist())

        gradients = self._stack_slopes(main_loop_frame['Gradients'], slope_axis=1)
        reference = self._stack_slopes(fits.getdata(path / 'Acq.DET1.REFSLP_WITH_OFFSETS_0001.fits'), slope_axis=1)[0]
        wfs = aotpy.ShackHartmann(
            uid='WFS',
            source=ngs,
            n_valid_subapertures=12,
            measurements=aotpy.Image('Gradients', gradients, time=loop_time),
            ref_measurements=aotpy.Image('Acq.DET1.REFSLP_WITH_OFFSETS', reference),
            subaperture_intensities=aotpy.Image(f'Intensities', main_loop_frame['Intensities'], time=loop_time)
        )

        wfs.non_common_path_aberration = aotpy.Aberration(
            uid='NCPA',
            modes=control_modes,
            coefficients=image_from_file(path / 'Ctr.MODAL_OFFSETS_ROTATED_0001.fits')  # in DM modal space
        )

        wfs.detector = aotpy.Detector(
            uid='DET',
            weight_map=image_from_file(path / 'Acq.DET1.WEIGHT_0001.fits'),
            dark=image_from_file(path / 'Acq.DET1.DARK_0001.fits'),
            flat_field=image_from_file(path / 'Acq.DET1.FLAT_0001.fits'),
            bad_pixel_map=image_from_file(path / 'Acq.DET1.DEAD_0001.fits'),
            sky_background=image_from_file(path / 'Acq.DET1.BACKGROUND_0001.fits')
        )

        pix_loop_frame = fits.getdata(path / 'NAOMI_PIXELS_0001.fits')
        wfs.detector.pixel_intensities = aotpy.Image(
            'Pixels',
            data=self._get_pixel_data_from_table(pix_loop_frame),
            time=aotpy.Time('Pixel time', frame_numbers=pix_loop_frame['FrameCounter'].tolist())
        )

        dm = aotpy.DeformableMirror('DM', telescope=self.system.main_telescope, n_valid_actuators=241)

        modal_coefficients = main_loop_frame['ModalCoefficients']
        modal_coefficients += fits.getdata(path / 'Ctr.MODAL_OFFSETS_ROTATED_0001.fits') * 2
        # These are saved in the DM modal space. Need to add the rotated offsets to get the real coefficients that are
        # then sent to the DM after M2DM conversion.
        s2m = self._stack_slopes(fits.getdata(path / 'Recn.REC1.CM_0001.fits'), slope_axis=1)
        # The S2M matrix is already rotated to to DM modes
        m2s = self._stack_slopes(fits.getdata(path / 'ModalRecnCalibrat.REF_IM_0001.fits'), slope_axis=0)

        try:
            ref_commands = aotpy.Image('Ctr.ACT_POS_REF_MAP', fits.getdata(path / 'Ctr.ACT_POS_REF_MAP_0001.fits')[0])
        except FileNotFoundError:
            ref_commands = None
            warnings.warn("Reference commands file not found ('Ctr.ACT_POS_REF_MAP_0001.fits').")

        loop = aotpy.ControlLoop(
            uid='Main Loop',
            input_sensor=wfs,
            commanded_corrector=dm,
            time=loop_time,
            time_filter_num=aotpy.Image('Ctr.TERM_A', fits.getdata(path / 'Ctr.TERM_A_0001.fits')),
            time_filter_den=aotpy.Image('Ctr.TERM_B', fits.getdata(path / 'Ctr.TERM_B_0001.fits')),
            commands=aotpy.Image('DM positions', main_loop_frame['Positions'], time=loop_time),
            ref_commands=ref_commands,
            modes=control_modes,
            modal_coefficients=aotpy.Image('Modal Coefficients', modal_coefficients, time=loop_time),
            measurements_to_modes=aotpy.Image('Recn.REC1.CM', s2m),
            modes_to_commands=image_from_file(path / 'RTC.M2DM_SCALED_0001.fits'),
            commands_to_modes=image_from_file(path / 'RTC.DM2M_SCALED_0001.fits'),
            modes_to_measurements=aotpy.Image('ModalRecnCalibrat.REF_IM', m2s),
            closed=main_hdr['ESO AOS LOOP ST'],
            framerate=main_hdr['ESO AOS LOOP RATE']
        )

        asm = aotpy.AtmosphericParameters(
            'ESO ASM (Astronomical Site Monitor)',
            wavelength=500e-9,
            fwhm=[main_hdr['ESO TEL AMBI FWHM']],
            tau0=[main_hdr['ESO TEL AMBI TAU0']],
            theta0=[main_hdr['ESO TEL AMBI THETA0']],
            layers_wind_direction=aotpy.Image('ESO TEL AMBI WINDDIR', np.array([[main_hdr['ESO TEL AMBI WINDDIR']]])),
            layers_wind_speed=aotpy.Image('ESO TEL AMBI WINDSP', np.array([[main_hdr['ESO TEL AMBI WINDSP']]]))
        )

        self.system.sources = [ngs]
        self.system.wavefront_sensors = [wfs]
        self.system.wavefront_correctors = [dm]
        self.system.loops = [loop]
        self.system.atmosphere_params = [asm]

    def _get_eso_telescope_name(self) -> str:
        # Allow for both:
        #   ESO-VLTI-Amnop
        #   ESO-VLTI-Uijkl-Amnop
        # as long as mnop contains the correct number
        return f"ESO-VLTI-%A%{self._at_number}%"

    def _get_eso_ao_name(self) -> str:
        return 'NAOMI'

    def _get_run_id(self) -> str:
        return '60.A-9278(D)'



"""
This module contains a classes used as a base for the rest of the core of aotpy.
"""

from collections import namedtuple
from dataclasses import dataclass
from typing import Any

__all__ = ['Referenceable', 'Coordinates', 'Metadatum']

Coordinates = namedtuple('Coordinates', 'x y')


@dataclass
class Metadatum:
    """Contains a set of key, value and (optionally) comment which describe the Image data in some aspect."""
    key: str
    "Identifier of the value."

    value: Any
    "Information that describes the Image data."

    comment: str = None
    "Optional description of the metadatum."

    def to_tuple(self) -> tuple:
        """ Return the `Metadatum` object expressed as a tuple (key, value, comment)."""
        return self.key, self.value, self.comment


@dataclass
class Referenceable:
    """Abstract class for all classes which can be referenced via a UID."""
    uid: str
    """Unique identifier of the object, which allows unambiguous referencing."""

    def __post_init__(self):
        if self.__class__ == Referenceable:
            raise TypeError("Cannot instantiate abstract class.")
